//-----------------------------LICENSE NOTICE--------------------------------
//  This file is part of CPCRogue: An Amstrad CPC rogue like game
//  Copyright (C) 2019 Andrés Mata Bretón (@FlautinesMata)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------
#include <cpctelera.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fighter.h"
#include "consts.h"
#include "entity.h"
#include "conio.h"
#include "aux_math.h"

extern u8 msg[];
///< Count of how many fighters we allocated
u8 num_fighters;
///< Array of fighter components so we use them as we need them
struct TFighter fighters[MAX_FIGHTER_COMPONENTS];

///< Mob default stats
const struct TFighterInfo mob_info[] = {
  //  HP, XP, STR, DEF, REF, DEX, MIN_DMG, MAX_DMG
    {  8, 25, 11,  12,  11,   11,   2,      6 },    // GOBLIN
    {  7, 30,  9,  10,  13,   11,   2,      4 },    // BAT
    { 10, 25, 10,  10,  13,   12,   2,      4 },    // SNAKE
    { 20, 30, 12,  10,  11,   11,   2,      6 },    // KOBOLD
    { 25, 50, 14,  12,   9,    9,   4,      8 },    // ORC
};

///< % Hit
const u8 hit_probability[] = {
128, 131, 134, 138, 141, 145, 148, 152, 155, 159, 162, 166, 169, 173, 176, 180, 183, 187, 190, 194, 197, 201, 204, 208, 211, 214, 218, 221, 225, 228, 232
};

///< Atack/Defense Bonus
const i8 bon_atkdef[] = {
-7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 9, 9, 10, 11, 12, 12, 13, 14, 15
};

// Forward declaration of private functions
bool is_a_hit(struct TFighter *attacker, struct TFighter *target);
u8 get_damage(struct TFighter *attacker, struct TFighter *target);


//---------------------------------------------------------------------------
void init_fighters()
{
  cpct_memset (fighters, 0, sizeof (fighters));
  num_fighters = 0;
}
//---------------------------------------------------------------------------
/*!
 */
struct TFighter *fighter_create (
  i8  hp,
  u16 xp,
  u8  strn,
  u8  def,
  u8  ref,
  u8  dex,
  u8  min_dmg, u8 max_dmg,
  void (*death_fn) (struct TEntity *))
{
  struct TFighter *fighter = NULL;

  if (num_fighters < MAX_FIGHTER_COMPONENTS)
  {
    // Get an empty slot from the fighters array
    fighter = &fighters[num_fighters++];

    // Set the combat attributes
    fighter->max_hp = fighter->hp = hp;
    fighter->xp     = xp;
    fighter->strn   = strn;
    fighter->def    = def;
    fighter->ref    = ref;
    fighter->dex    = dex;

    // Min/Max base damage
    fighter->dmg.min  = min_dmg;
    fighter->dmg.max  = max_dmg;

    // Set the death fn. to call when the entity dies
    fighter->death_fn = death_fn;

  }
  // Return the fighter ptr as our new allocated fighter
  return fighter;
}
//---------------------------------------------------------------------------
/*!
 *
 * A fighter component is responsible of attacking actions. It will use the
 * fighter component stats to determine damage and other data needed for an
 * attack. Note that an attack is performed between a *fighter component*
 * and a target *entity*.
 *
 *  \param fighter  Fighter component that will perform the attack
 *  \param target   Entity that will be attacked
 */
void fighter_attack (struct TFighter *fighter,
                     struct TEntity *target)
{
  u8 dmg = 0;

  sprintf(msg,"\477\10\13%s\477\11\10 attacks \477\10\11%s",
    name(fighter->owner), name(target));
  log_msg_ext (msg);

  if (is_a_hit (fighter, target->fighter)) {
    dmg = get_damage(fighter, target->fighter);
    fighter_take_damage (target->fighter, fighter->owner, dmg);
  }
  else {
    log_msg_ext ("\477\11\10but fails misserably");
  }
}
//---------------------------------------------------------------------------
/*!
 *
 */
void fighter_take_damage (struct TFighter *fighter,
                          struct TEntity *attacker,
                          i8 dmg)
{
  extern struct TEntity *player;
  extern bool stats_changed;

  fighter->hp -= dmg;

  sprintf (msg, "\477\11\10%s HP: \477\10\11%d\477\11\10/%d",
    name(fighter->owner), fighter->hp, fighter->max_hp);
  log_msg_ext (msg);

  // Update stats if player took damage
  if (fighter->owner == player)
    stats_changed = true;

  // kill entity if hp <= 0
  if (fighter->hp <= 0) {

    // yield xp to player for killing this mob
    if (fighter->owner != player && attacker == player) {
      player->fighter->xp += fighter->xp;
      check_level();
      stats_changed = true;
    }

    // Execute the death action
    if (fighter->death_fn) {
      fighter->death_fn (fighter->owner);
    }
  }
}
//---------------------------------------------------------------------------
/*!
 * On death, most mobs stop moving.
 */
void kill_mob (struct TEntity *mob) {

  sprintf (msg, "\477\11\10%s \477\11\12is dead!", name(mob));
  log_msg_ext (msg);

  mob->fighter  = NULL;
  mob->ai       = NULL;
  set_dead (mob);
}
//---------------------------------------------------------------------------
void kill_player(struct TEntity *e)
{
  extern bool game_running;
  game_running = false;
}


//---------------------------------------------------------------------------
//    PRIVATE FUNCTIONS
//---------------------------------------------------------------------------
bool is_a_hit(struct TFighter *attacker, struct TFighter *target)
{
  u8 hit_idx = attacker->dex - target->ref;
  u8 hit_chance = 0;
  u8 rnd_value = cpct_rand ();

  // get % hit from table
  hit_idx = hit_idx < 128 ? hit_idx : 0;
  hit_chance = hit_probability[hit_idx];

/*
  sprintf (msg, "\477\11\10%d [\477\13\10%d]",
    rnd_value, hit_chance);
  log_msg_ext (msg);
*/
  return (rnd_value < hit_chance);
}

//---------------------------------------------------------------------------
// Damage = [Weapon Damage] +
//          BonAtkDef [STR_attacker] - BonAtkDef [DEF_target]
//
// For now, let's make [Weapon Damage] = rand[min_dmg, max_dmg]
//
u8 get_damage(struct TFighter *attacker, struct TFighter *target)
{
  i8 bon_atk = bon_atkdef[attacker->strn];
  i8 bon_def = bon_atkdef[target->def];
  u8 weapon_damage = rand_range (attacker->dmg.min, attacker->dmg.max);
  u8 dmg = weapon_damage + bon_atk - bon_def;
  dmg = (dmg < 128) ? dmg : 0;

  sprintf (msg, "\477\11\10%d + %d - %d = \477\13\10%d",
    weapon_damage, bon_atk, bon_def, dmg);
  log_msg_ext (msg);

  return dmg;
}