//-----------------------------LICENSE NOTICE--------------------------------
//  This file is part of CPCRogue: An Amstrad CPC rogue like game
//  Copyright (C) 2019 Andrés Mata Bretón (@FlautinesMata)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------
#include <cpctelera.h>
#include <stdbool.h>
#include <stdio.h>
#include "consts.h"
#include "draw.h"
#include "entity.h"
#include "components/fighter.h"
#include "components/items.h"
#include "components/rect.h"
#include "components/container.h"
#include "game_map.h"
#include "conio.h"
#include "aux_math.h"
#ifdef DEBUG
  #include <assert.h>
#endif

#ifdef DEBUG
  // forward declaration private fn.
  void draw_axis();
#endif

//---------------------------------------------------------------------------
void draw_game(bool needs_redraw) {

  extern bool stats_changed;
  extern struct TEntity *player;
  struct TEntity **pentity;
  struct TEntity *e;

  //cpct_clearScreen_f64(0x0000);
  if (needs_redraw) {
    // Draw map
    draw_map();

    // Draw all entities
    pentity = last_entity_in_level;

    while (pentity-- != entities_in_level) {
      e = *pentity;
      if (in_world(e))
        if (distance_sqr (player->x,player->y, e->x, e->y) < FOV_R2)
          entity_draw (e);
    } // while

    // Draw the player
    //entity_erase (player);
    entity_draw (player);

  } // if needs_redraw

  // Draw stats if needed
  if (stats_changed) {
    clr_stats();
    draw_stats();
    stats_changed = false;
  }

}

//---------------------------------------------------------------------------
void draw_map() {
  extern struct TEntity *player;
  i8 x,y;
  i8 start_x, start_y;
  i8 end_x, end_y;

  start_x = player->x < FOV_RADIUS ? 0 : player->x - FOV_RADIUS;
  start_y = player->y < FOV_RADIUS ? 0 : player->y - FOV_RADIUS;
  end_x   = player->x >= MAP_WIDTH-FOV_RADIUS-1 ?
    MAP_WIDTH-1 : player->x+FOV_RADIUS;
  end_y   = player->y >= MAP_HEIGHT-FOV_RADIUS-1 ?
    MAP_HEIGHT-1 : player->y+FOV_RADIUS;

  for (y=start_y; y <= end_y; ++y)
  for (x=start_x; x <=end_x; ++x) {
    if ( !(player->x == x && player->y == y) )
      draw_tile (x, y);
  }
#ifdef DEBUG
  draw_axis();
#endif
}

//---------------------------------------------------------------------------
void draw_tile (i8 x, i8 y) {

  u8 fg_color = WALL_FG, bg_color = DEFAULT_BG;
  struct TTile *current_tile = &game_map.tiles[y][x];

#ifdef DEBUG
  assert (x >= 0 && y >= 0 && x < MAP_WIDTH && y < MAP_HEIGHT);
#endif

  // Draw tile
  if (current_tile->spr)
    putchar_f ((void*)VMEM_MAP, x,y, current_tile->spr, fg_color, bg_color);
}

//---------------------------------------------------------------------------
void entity_draw(struct TEntity *e) {
  putchar_f ((void*)VMEM_MAP, e->x,e->y, ech(e), 1, DEFAULT_BG);
}

//---------------------------------------------------------------------------
void draw_player_hp() {
  extern u8 msg[];
  extern struct TEntity *player;

  sprintf (msg, "HP: %d/%d ", player->fighter->hp, player->fighter->max_hp);
  cpct_drawStringM1_f (msg, (void*)(VMEM_STATS_HP),
    STATS_FG, DEFAULT_BG);
}

//---------------------------------------------------------------------------
void draw_player_strn()
{
  extern u8 msg[];
  extern struct TEntity *player;
  sprintf (msg, "STR: %d", player->fighter->strn);
  cpct_drawStringM1_f (msg, (void*)(VMEM_STATS_STR),
    STATS_FG, DEFAULT_BG);
}
//---------------------------------------------------------------------------
void draw_player_def()
{
  extern u8 msg[];
  extern struct TEntity *player;
  sprintf (msg, "DEF: %d", player->fighter->def);
  cpct_drawStringM1_f (msg, (void*)(VMEM_STATS_DEF),
    STATS_FG, DEFAULT_BG);
}
//---------------------------------------------------------------------------
void draw_player_ref()
{
  extern u8 msg[];
  extern struct TEntity *player;
  sprintf (msg, "REF: %d", player->fighter->ref);
  cpct_drawStringM1_f (msg, (void*)(VMEM_STATS_REF),
    STATS_FG, DEFAULT_BG);
}
//---------------------------------------------------------------------------
void draw_player_dex()
{
  extern u8 msg[];
  extern struct TEntity *player;
  sprintf (msg, "DEX: %d", player->fighter->dex);
  cpct_drawStringM1_f (msg, (void*)(VMEM_STATS_DEX),
    STATS_FG, DEFAULT_BG);
}

//---------------------------------------------------------------------------
void draw_player_hungry(bool clear)
{
  extern struct TEntity *player;
  cpct_drawStringM1_f (" HUNGRY ", (void*)(VMEM_HUNGRY),
    clear ? DEFAULT_BG : WARNING_FG, clear ? DEFAULT_BG : WARNING_BG);
}

//---------------------------------------------------------------------------
void draw_stats() {
  extern u8 msg[];
  extern struct TEntity *player;
  extern u8 dungeon_lvl;

  sprintf (msg, "Lvl %d/%d",
    player->xp_level, player->fighter->xp);
  cpct_drawStringM1_f (msg, (void*)(VMEM_STATS_XP),
    STATS_FG, DEFAULT_BG);

  draw_player_hp();
  draw_player_strn();
  draw_player_def();
  draw_player_ref();
  draw_player_dex();

  sprintf (msg, "Gold: %d", player->item->value);
  cpct_drawStringM1_f (msg, (void*)(VMEM_STATS_GOLD),
    STATS_FG, DEFAULT_BG);

  sprintf (msg, "DLvl: %d", dungeon_lvl);
  cpct_drawStringM1_f (msg, (void*)(VMEM_DUNGEON_LVL),
    STATS_FG, DEFAULT_BG);
}

//---------------------------------------------------------------------------
void draw_inventory()
{
  extern struct TEntity *player;
  extern u8 msg[];
  clr_log();

  if (!player->container->num_items) {
    sprintf (msg, "\477\11\12Inventory empty");
    log_msg_ext (msg);
  }
  else for (u8 i=0; i<player->container->num_items; ++i) {
    sprintf (msg, "\477\13\10%c\477\11\10 - %s",
      'a'+i,
      name(player->container->inventory[i]));

    log_msg_ext (msg);
  }
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#ifdef DEBUG
/*
void draw_vismap () {
  u8 x,y;

  cpct_clearScreen_f64(0x0000);
  draw_axis();

  y = MAP_HEIGHT;
  do {
    --y;
    x = MAP_WIDTH;
    do {
      --x;
      if (is_visible (x,y))
        putchar_f ((void*)VMEM_MAP, x, y, '.', PEN_BRIGHT, PEN_CLEAR);
      else
        putchar_f ((void*)VMEM_MAP, x, y, '#', PEN_BRIGHT, PEN_CLEAR);
    } while (x);
  } while (y);
}
*/
void draw_axis() {
  u8 y = MAP_HEIGHT;

  cpct_drawStringM1_f(" 0123456789012345678901234567",
    (void*)VMEM, PEN_BRIGHT, PEN_CLEAR);

  do {
    putchar_f ((void*)VMEM, 0,y, ((y-1)%10)+48, PEN_BRIGHT, PEN_CLEAR);
    --y;
  } while (y);
}

#endif