//-----------------------------LICENSE NOTICE--------------------------------
//  This file is part of CPCRogue: An Amstrad CPC rogue like game
//  Copyright (C) 2019 Andrés Mata Bretón (@FlautinesMata)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------
#include <cpctelera.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include "consts.h"
#include "entity.h"
#include "components/fighter.h"
#include "components/ai.h"
#include "components/container.h"
#include "components/items.h"
#include "components/rect.h"
#include "components/stairs.h"
#include "game_map.h"
#include "conio.h"
#include "draw.h"
#include "aux_math.h"

#ifdef DEBUG
  #include <assert.h>
#endif
// List of entities in our game
struct TEntity entities[MAX_ENTITIES];  // List of entities in game
u8 num_entities;

// Entity information
const struct TEntityInfo entity_info[] = {
  {PLAYER,        "",                 SPR_PLAYER},
// MONSTERS
  {GOBLIN,        "Goblin",           SPR_GOBLIN},
  {BAT,           "Bat",              SPR_BAT},
  {SNAKE,         "Snake",            SPR_SNAKE},
  {KOBOLD,        "Kobold",           SPR_KOBOLD},
  {ORC,           "Orc",              SPR_ORC},
// POTIONS
  {HEAL_POTION,   "Healing Potion",   SPR_POTION},
  {POISON_POTION, "Poison Potion",    SPR_POTION},
  {STR_POTION,    "Strength Potion",  SPR_POTION},
  {DEF_POTION,    "Defense Potion",   SPR_POTION},
  {DEX_POTION,    "Dexterity Potion", SPR_POTION},
  {REF_POTION,    "Reflexes Potion",  SPR_POTION},
// FOOD
  {FOOD,          "Food",             SPR_FOOD},
// GOLD
  {GOLD,          "Gold",             SPR_GOLD},
// ORUTRIX
  {ORUTRIX,       "The Orutrix",      SPR_ORUTRIX},
// STAIRS
  {STAIRSUP,      "Stairs Up",        SPR_STAIRSUP},
  {STAIRSDOWN,    "Stairs Down",      SPR_STAIRSDOWN},
};

//---------------------------------------------------------------------------
void init_entities() {
  cpct_memset (entities, 0x00, sizeof (entities));
  num_entities = 0;
}

//---------------------------------------------------------------------------
//void entity_erase (struct TEntity *e) {
//  draw_tile (e->px, e->py, PEN_BRIGHT, PEN_CLEAR);
//}

//---------------------------------------------------------------------------
struct TEntity *entity_create (enum Etype type, i8 x, i8 y,
  struct TFighter   *fighter,
  struct TAI        *ai,
  struct TContainer *container,
  struct TItem      *item,
  struct TStairs    *stairs)
{
  struct TEntity *e = NULL;

// If DEBUG ensure we didn't reached # entities limit
#ifdef DEBUG
  assert (num_entities < MAX_ENTITIES);
#endif

  // Get a ptr to next available slot in the entities array
  e = &entities[num_entities++];

  e->type = type;
  // A new created entity is assumed to be in the world
  add_to_world (e, x,y);

  if (fighter)                      // Set the owner of the component
    fighter->owner = e;             // Fighter component if it's a combat
  e->fighter = fighter;             // entity or NULL

  if (ai)                           // AI component
    ai->owner = e;
  e->ai = ai;

  if (container) {                  // Set owner of the container and
    init_container (container);     // ...initialize it
    container->owner = e;           //
  }
  e->container = container;         // Container if can collect items or NULL

  if (item)                         // Item component
    item->owner = e;
  e->item = item;

  if (stairs)                       // Stairs component
    stairs->owner = e;
  e->stairs = stairs;

  return e;
}

//---------------------------------------------------------------------------
bool entity_move (struct TEntity *e, i8 dx, i8 dy)  {

  struct TEntity *target = NULL; // Used when searching entities

  // If the tile @ new position blocks movement (i.e: wall) do nothing
  if (!tile_blocks_movement (e->x+dx, e->y+dy)) {

    // ATTACK if there's a fighter entity there
    if ((target = get_fighter_at(e->x+dx, e->y+dy) )) {
#ifdef DEBUG
      assert (target);
#endif
      if (target != e)  // Ensure we are not attacking ourselves!
        // Attack target
        fighter_attack (e->fighter, target);
    } // if (get_fighter_at)
    else {
      // No blocking entity, MOVE to the new position
      e->px = e->x;     // Save old positions
      e->py = e->y;
      e->x  = e->x+dx;  // Update to new position
      e->y  = e->y+dy;
    } // else (get_fighter_at)

    // entity moved, return true
    return true;
  } // if (!tile_blocks_movement)

  // entity did not move, return false
  return false;
}

bool is_there_an_entity (i8 x, i8 y)
{
  struct TEntity **pentity;
  pentity = last_entity_in_level;
  while(pentity-- != entities_in_level) {
    if ((*pentity) && (*pentity)->x == x && (*pentity)->y ==y)
      return true;
  }
  return false;
}
//---------------------------------------------------------------------------
/*! \brief
 * Gets the first blocking entity in a tile
 *
 * \param[in] x,y tile position to check for pressence of entity
 */
struct TEntity *get_fighter_at (i8 x, i8 y)
{
  struct TEntity **pentity;
  pentity = last_entity_in_level;
  while(pentity-- >= entities_in_level) {
    if ((*pentity)->fighter &&
      (*pentity)->x == x && (*pentity)->y == y)
      return (*pentity);
  }
  return NULL;
}

//---------------------------------------------------------------------------
/*! \brief
 * Gets the first item entity in a tile
 *
 * \param[in] x,y tile position to check for pressence of entity
 */
struct TEntity *get_item_at (i8 x, i8 y)
{
  struct TEntity **pentity;
  pentity = last_entity_in_level;
  while(pentity-- != entities_in_level) {
    if (in_world((*pentity)) &&(*pentity)->item &&
     (*pentity)->x == x &&(*pentity)->y == y)
      return (*pentity);
  }

  return NULL;
}
//---------------------------------------------------------------------------
struct TEntity *get_stairs_at (i8 x, i8 y)
{
  struct TEntity **pentity;
  pentity = last_entity_in_level;
  while(pentity-- != entities_in_level) {
    if ((*pentity)->stairs &&
      (*pentity)->x == x && (*pentity)->y == y)
      return *pentity;
  }
  return NULL;
}

//---------------------------------------------------------------------------
void add_to_world (struct TEntity *e, i8 x, i8 y) {

#ifdef DEBUG
  assert (num_entities < MAX_ENTITIES-1);
#endif

  e->x = e->px = x;
  e->y = e->py = y;
  e->flags |= E_IN_WORLD;
  //num_entities++;
}

//---------------------------------------------------------------------------
void remove_from_world (struct TEntity *e) {

  e->flags &= (~E_IN_WORLD);
  //num_entities--;
}

//---------------------------------------------------------------------------
void entity_find (struct TEntity *array[], struct TEntity *e, u8 *index)
{
  struct TEntity **parray = array;

  for (*index = 0;  *parray != e; parray++, (*index)++)
    ;
}

//---------------------------------------------------------------------------
void remove_from_list (struct TEntity *array[], struct TEntity *e)
{
  //extern struct TEntity *entities_in_level[];
  u8 entity_idx;
  entity_find (entities_in_level, e, &entity_idx);
  --last_entity_in_level;
  array[entity_idx] = *last_entity_in_level;
}

void add_to_list (struct TEntity *array[], struct TEntity *e)
{
  *last_entity_in_level = e;
  last_entity_in_level++;
}

//---------------------------------------------------------------------------
void check_level()
{
  extern struct TEntity *player;
  u16 level_up_xp = LEVEL_UP_BASE + player->xp_level * LEVEL_UP_FACTOR;

  // See if the player's experience is enough to level up
  if (player->fighter->xp >= level_up_xp) {
    player->xp_level += 1;
    player->fighter->xp -= level_up_xp;
    player->fighter->max_hp += rand_range (player->xp_level+2, player->xp_level+8);
    player->fighter->hp = player->fighter->max_hp;

    // Improve stats
    // every 2 levels +STR
    if (!(player->xp_level & 2))  player->fighter->strn += 1;
    // every 4th levels +REF
    if (!(player->xp_level & 3))   player->fighter->ref += 1;
    // every 2th and 9th level +DEX
    if (!(player->xp_level & 5))   player->fighter->dex += 1;
    // 2nd,4th,6th,16th,18th,20th,22nd +DEF
    if (!(player->xp_level & 9))   player->fighter->def += 1;
  }
}

//---------------------------------------------------------------------------
char dead_name[25];
char *name(struct TEntity *e)
{
  if (!is_dead(e)) {
    if (is_potion(e) && !is_known(e))
      return "UNKNOWN POTION";
    else
    return entity_info[e->type].name;
  }
  else {
    sprintf(dead_name, "%s %s", "Dead", entity_info[e->type].name);
    return dead_name;
  }

}

//---------------------------------------------------------------------------
bool is_potion(struct TEntity *e)
{
  return ( (e->type >= HEAL_POTION) && (e->type < GOLD) );
}

//---------------------------------------------------------------------------
bool is_known(struct TEntity *e)
{
  extern u8 known_potions[];
  return known_potions[e->type-HEAL_POTION];
}

//---------------------------------------------------------------------------
void remove_bonus(struct TEntity *target)
{
  if (target->flags & E_STR_BONUS) {
    target->flags &= ~E_STR_BONUS;
    target->fighter->strn -= 2;
    draw_player_strn();
  }
  if (target->flags & E_REF_BONUS) {
    target->flags &= ~E_REF_BONUS;
    target->fighter->ref -= 2;
    draw_player_ref();
  }
  if (target->flags & E_DEX_BONUS) {
    target->flags &= ~E_DEX_BONUS;
    target->fighter->dex -= 2;
    draw_player_dex();
  }
  if (target->flags & E_DEF_BONUS) {
    target->flags &= ~E_DEF_BONUS;
    target->fighter->def -= 2;
    draw_player_def();
  }
}