//-----------------------------LICENSE NOTICE--------------------------------
//  This file is part of CPCRogue: An Amstrad CPC rogue like game
//  Copyright (C) 2019 Andrés Mata Bretón (@FlautinesMata)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------
#include <cpctelera.h>
#include <stdbool.h>
#include "consts.h"
#include "entity.h"
#include "components/rect.h"
#include "game_map.h"
#include "aux_math.h"
#include "generators.h"

#ifdef DEBUG
#include <assert.h>
#endif

extern struct TEntity entities[];
extern u8 num_entities;
extern struct TEntity *player;

// forward declaration of private functions
void  _init_tiles (u8 width, u8 height);
void  _place_rooms_in_grid();
void  _grid_to_map ();
void  _carve_room (struct TRoom *r);
void  _dig_tunnel_x (u8 src_x, u8 src_y, u8 dst_x);
void  _dig_tunnel_y (u8 src_x, u8 src_y, u8 dst_y);
void _place_stairs();
void _place_entities(struct TRoom *r);
void _place_orutrix();

// Instantiate the global variables used in this module
struct TGridCell  grid[GRID_SIZE];
struct TRoom      rooms[MAX_ROOMS];
struct TRoom      *plast_room;      ///< Pointer to last room created
/*!
 * The global variable **game_map** holds information about the tiles in the
 * current generated map. It's a global variable and can be accessed from
 * any file to query information about a specific tile or dimensions of the
 * map.
 */
struct TMap game_map;
struct TEntity *entities_in_level[MAX_ENTITIES];
struct TEntity **last_entity_in_level;

/*!
 * Creates a very basic dungeon
 *
 * \param[in] width Width of the map
 * \param[in] height Height of the map
 */
void map_create (u8 width, u8 height)
{
  //extern bool   orutrix_placed;
  extern u8     orutrix_dlvl;
  extern u8     dungeon_lvl;

  _init_tiles(width, height);
  _place_rooms_in_grid();
  _grid_to_map();
  _place_stairs();

  // Place orutrix if we are on the corresponding dungeon level
  if (orutrix_dlvl == dungeon_lvl) {
    _place_orutrix();
    //orutrix_placed = true;
  }
}
//--------------------------------------------------------------------------
void get_random_position (struct TRect *r, i8 *x, i8 *y)
{
  do {
    *x = rand_range (r->left+1, r->right-1);
    *y = rand_range (r->top+1, r->bottom-1);
  } while (tile_blocks_movement (*x,*y) || is_there_an_entity(*x,*y));
}
//===========================================================================
//===========================================================================
/***
 *       ___  ____ _ _  _ ____ ___ ____
 *       |__] |__/ | |  | |__|  |  |___
 *       |    |  \ |  \/  |  |  |  |___
 *    ____ _  _ _  _ ____ ___ _ ____ _  _ ____
 *    |___ |  | |\ | |     |  | |  | |\ | [__
 *    |    |__| | \| |___  |  | |__| | \| ___]
 *
 */
//===========================================================================
//===========================================================================

/*! \brief
 * Initializes the map to a given width and height.
 *
 * All the tiles in the map  are set to block movement and light. Also
 * initializes the visibility map to all cells not visible.
 *
 * \param width   Width of the map to initialize
 * \param height  Height of the map to initialize
 */
 void _init_tiles (u8 width, u8 height)
 {
  game_map.width  = width;
  game_map.height = height;

  // All tiles by default to block movement (walls)
  //cpct_memset (game_map.tiles, 0, sizeof (game_map.tiles));
  for (u8 y=0;y<height;y++) {
    for (u8 x=0;x<width;x++) {
      game_map.tiles[y][x].spr     = '\0';
      game_map.tiles[y][x].t_flags = BLOCKED;
    }
  }
 }
 //--------------------------------------------------------------------------
/*! \brief
 * Distributes rooms in a grid system
 *
 * Each room is placed in a single grid cell, thus we ensure no rooms
 * overlap. This fn. does the following:
 *
 *  - Zero all rooms data
 *  - Create a rnd # of rooms of rnd size
 *  - Set the top,left coordinate of each grid cell
 *  - Place a room in a rnd grid cell
 *
 * After this function is called each tile has a room id assigned to it.
 * But the room tiles are not yet "carved"on the map. That's the role of
 * \ref _grid_to_map which is called next.
 */
void _place_rooms_in_grid()
{
  struct TRoom *proom;     // Room to place
  u8 x,y;                 // x,y for the new room room
  u8 w,h;                 // width, height for the new room
  u8 num_rooms;           // How many rooms
  u8 i;
  u8 spot;                // Empty spot in the grid

  // Zero all rooms
  cpct_memset (rooms, 0, MAX_ROOMS * sizeof (struct TRoom));

  // How many rooms
  num_rooms = rand_range (MIN_ROOMS, MAX_ROOMS);
  plast_room = &rooms[num_rooms];

  // Build that many rooms
/*
  for (proom = rooms; proom != plast_room; proom++) {

    w = rand_range (MIN_ROOM_SIZE, MAX_ROOM_SIZE);
    h = rand_range (MIN_ROOM_SIZE, MAX_ROOM_SIZE);
    rect_create (&(proom->rect), 0,0, w, h);
  }
*/

  // Setup the grid coordinates of each cell
  for (i=0; i != GRID_SIZE; ++i) {
    x = (i%GRID_WIDTH) * GRID_CELL_WIDTH;
    y = (i/GRID_WIDTH) * GRID_CELL_HEIGHT;
    grid[i].x = x;
    grid[i].y = y;
    grid[i].room_id = ROOM_NONE;
  }

  // Add rooms to grid
  for (i=0, proom=rooms; proom != plast_room; ++i, ++proom) {
    // Find an empty spot on the grid
    do {
      spot = rand_range (0, GRID_SIZE);
    } while (grid[spot].room_id != ROOM_NONE);

    // Update x,y in room
    x = grid[spot].x;
    y = grid[spot].y;
    w = rand_range (MIN_ROOM_SIZE, MAX_ROOM_SIZE);
    h = rand_range (MIN_ROOM_SIZE, MAX_ROOM_SIZE);
    rect_create (&(proom->rect), x,y, w, h);
    grid[spot].room_id  = i;
#ifdef DEBUG
    proom->room_id = i;
#endif
  }
}
//---------------------------------------------------------------------------
/*! \brief
 * "Carve" rooms and connects them.
 */
void  _grid_to_map ()
{
  struct TRoom *proom;
  u8 i, new_x, new_y, prev_x, prev_y;
  i = new_x = new_y = prev_x = prev_y = 0;

  cpct_memset (entities_in_level, 0, sizeof (entities_in_level));
  last_entity_in_level = entities_in_level;

  // For each room create the tunnels that connect them
  for (i=0, proom=rooms; proom != plast_room; ++proom, ++i) {
    // Center coords of this room for later use
    rect_get_center (&(proom->rect), &new_x, &new_y);

    // dig tunnels after first room
    if (i != 0) {
      _dig_tunnel_x (prev_x, prev_y, new_x);
      _dig_tunnel_y (new_x,  prev_y, new_y);
    }
    else {
      // Add player to 1st room
      player->x = new_x;
      player->y = new_y;

      // Add player to entities in this level
      *last_entity_in_level++ = player;
    }
    // save this room coords. for digging
    prev_x = new_x;
    prev_y = new_y;
  }

  // All tunnels are created, now carve the rooms over the tunnels
  // Doing in that order makes it easy to "hide" the parts of the
  // tunnels crossing the rooms.
  // We also place the entities and items in each room
  for (i=0, proom=rooms; proom != plast_room; ++proom, ++i) {
    _carve_room (proom);
    _place_entities(proom);
  }
}
//--------------------------------------------------------------------------
/*! \brief
 *  Fill a room with floor tiles and set not blocking flags.
 *
 *  \param r Room pointer to carve
 *
 */
void  _carve_room (struct TRoom *r)
{
  u8 x,y;
  char spr;

  for (y = r->rect.top+1; y <= r->rect.bottom-1; ++y) {
    for (x = r->rect.left+1; x <= r->rect.right-1; ++x) {
      game_map.tiles[y][x].spr = SPR_FLOOR;
      game_map.tiles[y][x].t_flags &= ~BLOCKED;
    }
  }
  // vertical walls
  for (y=r->rect.top+1; y < r->rect.bottom; ++y) {
    if (tile_blocks_movement (r->rect.left, y))
      spr = SPR_VWALL;
    else
      spr = ' ';
    game_map.tiles[y][r->rect.left].spr =spr;

    if (tile_blocks_movement (r->rect.right, y))
      spr = SPR_VWALL;
    else
      spr = ' ';
    game_map.tiles[y][r->rect.right].spr =spr;
  }
  // horizontal walls
  for (x=r->rect.left+1; x < r->rect.right; ++x) {
    spr = (tile_blocks_movement (x, r->rect.top)) ? SPR_HWALL : ' ';
    game_map.tiles[r->rect.top][x].spr = spr;
    spr = (tile_blocks_movement (x, r->rect.bottom)) ? SPR_HWALL : ' ';
    game_map.tiles[r->rect.bottom][x].spr = spr;
  }
  // corners
  #ifdef DEBUG
  game_map.tiles[r->rect.top][r->rect.left].spr = '0'+r->room_id;
  #else
  game_map.tiles[r->rect.top][r->rect.left].spr = SPR_TLWALL;
  #endif
  game_map.tiles[r->rect.bottom][r->rect.left].spr = SPR_BLWALL;
  game_map.tiles[r->rect.top][r->rect.right].spr = SPR_TRWALL;
  game_map.tiles[r->rect.bottom][r->rect.right].spr = SPR_BRWALL;

}
//--------------------------------------------------------------------------
/*! \brief
 *  'Carves' a horizontal tunnel from source x,y to destination x
 *
 *  \param src_x  Source x tile
 *  \param src_y  Source y tile
 *  \param dst_x  Destination x tile (destination y = src_y)
 *
 */
void  _dig_tunnel_x (u8 src_x, u8 src_y, u8 dst_x)
{
  u8 start_x, end_x;
  struct TTile *tile;
  struct TTile *tile_end;

  start_x = min (src_x, dst_x);
  end_x   = max (src_x, dst_x);

  tile = &(game_map.tiles[src_y][start_x]);
  tile_end = tile + end_x - start_x;
  while (tile <= tile_end) {
    tile->spr      = SPR_TUNNEL;
    tile->t_flags &= ~BLOCKED;
    tile++;
  }
}
//--------------------------------------------------------------------------
/*! \brief
 *  'Carves' a vertical tunnel from source x,y to destination x
 *
 *  \param src_x  Source x tile
 *  \param src_y  Source y tile
 *  \param dst_x  Destination y tile (destination x = src_x)
 *
 */
void  _dig_tunnel_y (u8 src_x, u8 src_y, u8 dst_y)
{
  u8 start_y, end_y;
  struct TTile *tile;
  struct TTile *tile_end;
  bool out_of_room = false;

  start_y = min (src_y, dst_y);
  end_y   = max (src_y, dst_y);

  tile = &(game_map.tiles[start_y][src_x]);
  tile_end = tile + (end_y - start_y)*MAP_WIDTH;
  while (tile < tile_end) {
    tile->spr      = SPR_TUNNEL;
    tile->t_flags &= ~BLOCKED;
    tile+=MAP_WIDTH;
  }

}
//--------------------------------------------------------------------------
void _place_stairs()
{
  extern u8 orutrix_dlvl;
  extern u8 dungeon_lvl;
  struct TEntity **e = last_entity_in_level;
  struct TRoom *proom;
  u8 x,y;

  // Place stairs down in all levels but the last one where the Orutrix is.
  if (orutrix_dlvl != dungeon_lvl) {
    // Down stairs at 1st room
    proom = plast_room-1;
    rect_get_center (&(proom->rect), &x, &y);
    *e++ = gen_stairs (x, y, true);
    last_entity_in_level = e;
  }


  // Up stairs
  /*proom = rooms;
  rect_get_center (&(proom->rect), &x, &y);
  *e++ = gen_stairs (x, y, false);
  last_entity_in_level = e;
  */
}

//--------------------------------------------------------------------------
void _place_entities(struct TRoom *proom)
{
  u8 x,y;
  u8 number_of_mobs   = rand_range (0, MAX_MOBS_PER_ROOM);
  u8 number_of_items  = rand_range (0, MAX_ITEMS_PER_ROOM);

  for (u8 i=number_of_mobs; i!=0; i--) {
    get_random_position (&(proom->rect), &x,&y);
    *last_entity_in_level++ = gen_mob (ERANDOM, x,y);
  }

  for (u8 i=number_of_items; i!=0; i--) {
    get_random_position (&(proom->rect), &x,&y);
    *last_entity_in_level++ = gen_item (ERANDOM, x,y);
  }
}

//--------------------------------------------------------------------------
void _place_orutrix()
{
  struct TEntity **e = last_entity_in_level;
  struct TRoom *proom;
  u8 x,y;

  // place it in the last room
  proom = plast_room - 1;
  get_random_position (&(proom->rect), &x,&y);
  *e++ = gen_orutrix (x,y);
  last_entity_in_level = e;
}