//-----------------------------LICENSE NOTICE--------------------------------
//  This file is part of CPCRogue: An Amstrad CPC rogue like game
//  Copyright (C) 2019 Andrés Mata Bretón (@FlautinesMata)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------

/*!
 * \defgroup game_map
 *
 * \brief Map creation and query routines
 *
 * The structures and functions in this module helps with the creation and
 * query of the map in the game.
 *
 * The game stores information of each tile in the map. Each tile needs to
 * keep track of certain attributes like is it walkable?, or blocks light?
 * etc.
 *
 */
/**@{*/

/*!
 * \struct  TTile
 *
 * \brief Single tile information
 */
struct TTile {
  u8 spr;
  u8 t_flags; ///< flag attributes for a specific tile
};

/*!
 * \struct TMap
 *
 * \brief Map dimensions and tiles it contains
 *
 */
struct TMap{
  u8 width;   ///< Game map width
  u8 height;  ///< Game map height
  struct TTile tiles[MAP_HEIGHT][MAP_WIDTH]; ///<every single tile in the map
};

/*!
 * \struct TRoom
 *
 * \brief Rooms in the level.
 *
 * A room is a boxed empty space surrounded by walls. It is mainly used
 * internally to randomly generate the dungeon.
 *
 */
// Room info
typedef struct TRoom {
  struct TRect rect;  ///< x,y, width, height including walls
#ifdef DEBUG
  u8           room_id;
#endif
};

/*!
 * \struct TGridCell
 *
 * \brief Grid cell info, used to distribute the rooms
 *
 * The creation of the randomly generated dungeon uses a grid based
 * distribution to ensure no rooms will overlap. This struct keeps info
 * about this grid.
 *
 */
struct TGridCell {
  u8 x;       ///< left coordinate of the grid cell
  u8 y;       ///< top coordinate of the grid cell
  u8 room_id; ///< room assigned to this grid cell
};
extern struct TGridCell grid[];       ///< Grid cells
extern struct TRoom     rooms[];      ///< Rooms placed in map
extern struct TRoom     *plast_room;
/*!
 * \brief Global variable to access the game map
 */
extern struct TMap game_map;
extern struct TEntity *entities_in_level[];
extern struct TEntity **last_entity_in_level;

/*!
 * \brief Create a new map
 */
void map_create (u8 width, u8 height);

/*!
 * \brief True if specified map coordinates block movement (i.e: wall)
 *
 * \param[in] x,y Map coordinates to query.
 *
 * \returns true if tile blocks movement.
 */
inline bool tile_blocks_movement (i8 x, i8 y)
{
  return game_map.tiles[y][x].t_flags & BLOCKED;
}
/*!
 * \brief Gets a non blocking random position tile inside the coordinates
 * contained in a TRect area.
 *
 * \param[in]  TRect Rectangle to limit the random position.
 * \param[out] x,y Coordinates of a non blocking tile
 *
 */
void get_random_position (struct TRect *r, i8 *x, i8 *y);